@@ -33,6 +33,94 @@ static void sigchld(int fd, short event, void *ev_)
     _exit(EXIT_SUCCESS);
 }
 
+static void sigterm(int fd, short event, void *ev_)
+{
+    struct event *ev = ev_;
+
+    (void) fd;
+    (void) event;
+    if (expire_pid >= (pid_t) 1) {
+        (void) kill(expire_pid, SIGTERM);
+    }
+    signal_del(ev);
+}
+
+static int changeuidgid(void)
+{
+    if (uid == (uid_t) 0) {
+        return 0;
+    }
+    if (
+#ifdef HAVE_SETGROUPS
+        setgroups(1U, &gid) ||
+#endif
+        setgid(gid) || setegid(gid) ||
+        setuid(uid) || seteuid(uid) || chdir("/")) {
+        return -1;
+    }
+    return 0;
+}
+
+static void clearargs(int argc, char **argv)
+{
+#ifndef NO_PROCNAME_CHANGE
+# if defined(__linux__) && !defined(HAVE_SETPROCTITLE)
+    int i;
+
+    for (i = 0; environ[i] != NULL; i++);
+    argv0 = argv;
+    if (i > 0) {
+        argv_lth = environ[i-1] + strlen(environ[i-1]) - argv0[0];
+    } else {
+        argv_lth = argv0[argc-1] + strlen(argv0[argc-1]) - argv0[0];
+    }
+    if (environ != NULL) {
+        char **new_environ;
+        unsigned int env_nb = 0U;
+
+        while (environ[env_nb] != NULL) {
+            env_nb++;
+        }
+        if ((new_environ = malloc((1U + env_nb) * sizeof (char *))) == NULL) {
+            abort();
+        }
+        new_environ[env_nb] = NULL;
+        while (env_nb > 0U) {
+            env_nb--;
+            /* Can any bad thing happen if strdup() ever fails? */
+            new_environ[env_nb] = strdup(environ[env_nb]);
+        }
+        environ = new_environ;
+    }
+# else
+    (void) argc;
+    (void) argv;
+# endif
+#endif
+}
+
+static void setprogname(const char * const title)
+{
+#ifndef NO_PROCNAME_CHANGE
+# ifdef HAVE_SETPROCTITLE
+    setproctitle("-%s", title);
+# elif defined(__linux__)
+
+    if (argv0 != NULL) {
+        memset(argv0[0], 0, argv_lth);
+        strncpy(argv0[0], title, argv_lth - 2);
+        argv0[1] = NULL;
+    }
+# elif defined(__hpux__)
+    union pstun pst;
+    
+    pst.pst_command = title;
+    pstat(PSTAT_SETCMD, pst, strlen(title), 0, 0);
+# endif
+#endif
+    (void) title;
+}
+
 static void usage(void)
 {
     puts("\n" PACKAGE_STRING " - " __DATE__ "\n");
@@ -47,6 +135,8 @@ static void usage(void)
             "--timeout=xxx     (-t) disconnect after xxx seconds without data\n"
             "--debug           (-D) verbose debug mode\n"
             "--daemonize       (-B) operate in background\n"
+            "--uid=xxx         (-u) change user id\n"
+            "--gid=xxx         (-g) change group id\n"
             "\n"
             "Report bugs and suggestions to "
             ), stdout);
@@ -60,6 +150,7 @@ int main(int argc, char *argv[])
     struct addrinfo hints, *res;
     struct event ev;
     struct event sigchld_ev;
+    struct event sigterm_ev;
     int on;    
     int listen_fd;
     int fodder;
@@ -151,7 +242,32 @@ int main(int argc, char *argv[])
             if ((storage_dir = strdup(optarg)) == NULL) {
                 die_mem();
             }
-            strlen_storage_dir = strlen(storage_dir);
+            break;
+        }
+        case 'g': {
+            const char *nptr;
+            char *endptr;
+            
+            nptr = optarg;
+            endptr = NULL;
+            gid = (gid_t) strtoul(nptr, &endptr, 10);
+            if (!nptr || !*nptr || !endptr || *endptr) {
+                logfile(LOG_ERR, _("Illegal GID - Must be a number\n"));
+                return 1;
+            }
+            break;            
+        }
+        case 'u': {
+            const char *nptr;
+            char *endptr;
+            
+            nptr = optarg;
+            endptr = NULL;
+            uid = (uid_t) strtoul(nptr, &endptr, 10);
+            if (!nptr || !*nptr || !endptr || *endptr) {
+                logfile(LOG_ERR, _("Illegal UID - Must be a number"));
+                return 1;
+            }
             break;
         }
         default: {
@@ -199,7 +315,15 @@ int main(int argc, char *argv[])
     }
     freeaddrinfo(res);
 
+    if (chroot(storage_dir) != 0 || chdir("/") != 0) {
+        logfile(LOG_ERR, _("Unable to chroot to [%s]: [%s]"), storage_dir,
+                strerror(errno));
+        return 1;
+    }
+    changeuidgid();
+    (void) umask((mode_t) 0);
     dodaemonize();
+    clearargs(argc, argv);
     
     if ((expire_pid = fork()) == (pid_t) -1) {
         logfile(LOG_ERR, _("Unable to fork: [%s]"), strerror(errno));
@@ -208,12 +332,16 @@ int main(int argc, char *argv[])
     }
     if (expire_pid == (pid_t) 0) {
         (void) close(listen_fd);
+        setprogname("sharedanced [CLEANUP]");
         expire();
         _exit(0);
-    }        
+    }
+    setprogname("sharedanced [SERVER]");
     event_init();
     signal_set(&sigchld_ev, SIGCHLD, sigchld, &ev);
     signal_add(&sigchld_ev, NULL);
+    signal_set(&sigterm_ev, SIGTERM, sigterm, &ev);
+    signal_add(&sigterm_ev, NULL);
     event_set(&ev, listen_fd, EV_READ | EV_PERSIST, new_client, &ev);
     event_add(&ev, NULL);
     event_dispatch();
